// C++ Tracing implementation file
//
// Copyright (c) David Drysdale 2000
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
// AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
// AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// Except as contained in this notice, the name of the author shall not be
// used in advertising or otherwise to promote the sale, use or other dealings
// in this Software without prior written authorization from the author.

#include "cpptrace.h"
#include <tracecfg.h>

#ifdef TRC_COMPILE_LEVEL
#if (TRC_COMPILE_LEVEL <= MAX_TRACE_LEVEL)

#ifdef _WIN32
// -------------------------WIN32--------------------------------
#include <process.h>
#define UMASK(U)
#define GETPID   _getpid
#define TRACEFILE1 TRACE_FILENAME1
#define TRACEFILE2 TRACE_FILENAME2
#define LOCKFILE   TRACE_LOCK_FILENAME
#define SHMFILE    TRACE_SHM_FILENAME
#else
// -------------------------UNIX---------------------------------
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <unistd.h>
#define UMASK(U) umask(U)
#define GETPID   getpid
#define TRACEFILE1 TRACE_DIR_UNIX TRACE_FILENAME1
#define TRACEFILE2 TRACE_DIR_UNIX TRACE_FILENAME2
#define LOCKFILE   TRACE_DIR_UNIX TRACE_LOCK_FILENAME
#define SHMFILE    TRACE_DIR_UNIX TRACE_SHM_FILENAME
// --------------------------------------------------------------
#endif

#include <cstdlib>
#include <cstring>

static const int MAJORVERSION = 1;
static const int MINORVERSION = 0;

// Per-process data which is accessed outside this file
TraceOverall* TraceOverall::spInstance(0);
std::ofstream* TraceOverall::spTraceStream(0);
int TraceOverall::sPid(0);
int TraceOverall::sLockFD(-1);
bool TraceOverall::sInTrace(false);

#ifdef MT_TRACE
#include <lock.h>
// Instantiate a class whose constructor creates the critical section object
// used for locking multi-threaded tracing.  The destructor removes this
// critical section object; the code that uses this critical section object
// has to watch out for the object disappearing under it (ie it needs to
// test that (*sppTraceThreadLock) is non-null).  If this object gets
// destroyed while there are still >1 threads around, then there may be
// problems.
class InitOnlyLock {
 public:
  InitOnlyLock() {mpCSO = new CriticalSectionObject();}
  ~InitOnlyLock() {delete mpCSO; mpCSO = 0;}
  CriticalSectionObject *mpCSO;
};
InitOnlyLock sInitOnlyTraceLock;
CriticalSectionObject** sppTraceThreadLock = &(sInitOnlyTraceLock.mpCSO);
#endif

// Per-process data which only needs to be accessed in this file
static int sFileId = -1;
static void * spSharedMem = 0;

// Class for setting a umask within a scope, exception safely
class Umask {
public:
  explicit Umask(mode_t newMask) { mPrev=UMASK(newMask);}
  ~Umask() {UMASK(mPrev);}
private:
  mode_t mPrev;
};

TraceOverall* TraceOverall::instance()
{
  // Check whether this is the first entry for this process
  if (spInstance == 0) {
    // Zero out the umask temporarily so all can access trace files
    Umask tempUmask(0);

    // Create/open the trace lock file, which locks access to the shared data
    sLockFD = open(LOCKFILE, O_WRONLY | O_CREAT | O_APPEND, 0666);
    if (sLockFD == -1) throw std::string("failed to open trace lock file");

    // lock for the duration of this scope
    TraceLock scopeLock(sLockFD);

    // Get access to the trace shared memory
    bool firstin = false;
    sFileId = open(SHMFILE, O_RDWR, 0666);
    if (sFileId == -1) {
      // failed to access shared memory file => we are first in, so create
      // it and force it to be big enough
      firstin = true;
      sFileId = open(SHMFILE, O_RDWR|O_CREAT|O_TRUNC, 0666);
      if (sFileId == -1) throw std::string("failed to create shared memory file");
      if (lseek(sFileId, sizeof(TraceOverall)-1, SEEK_SET) == -1) throw std::string("failed to move to end of file");
      if (write(sFileId, "", 1) != 1) throw std::string("failed to write final byte");
    }

    // Map the shared memory file into memory
    spSharedMem = mmap(0, sizeof(TraceOverall), PROT_READ|PROT_WRITE, MAP_SHARED, sFileId, 0);
    if (spSharedMem == reinterpret_cast<void*>(-1)) throw std::string("failed to memory map the file");

#ifdef MT_TRACE
    bool showTid = true;
#else
    bool showTid = false;
#endif

    if (firstin) {
      // Put the single instance of the TraceOverall class for this process into the
      // shared memory area
      spInstance = new(spSharedMem) TraceOverall(TRACEFILE1,
                                                 TRACEFILE2,
                                                 TRACE_DEFAULT_LEVEL,
                                                 showTid,
                                                 TRACE_DEFAULT_CLASSWIDTH,
                                                 TRACE_DEFAULT_METHODWIDTH,
                                                 TRACE_DEFAULT_LINENOWIDTH,
                                                 TRACE_DEFAULT_MAXSIZE);
    } else {
      // There is already a TraceOverall object in the shared memory, so use it.
      spInstance = static_cast<TraceOverall*>(spSharedMem);

      // Check the version number and size
      if ((spInstance->mMajorVersion != MAJORVERSION) ||
          (spInstance->mStructSize != sizeof(TraceOverall))) {
        if (spInstance->mUseCount == 0) {
          // If the existing trace info is incompatible, but not in use, then we can
          // just trash it with our own
          spInstance = new(spSharedMem) TraceOverall(TRACEFILE1,
                                                     TRACEFILE2,
                                                     TRACE_DEFAULT_LEVEL,
                                                     showTid,
                                                     TRACE_DEFAULT_CLASSWIDTH,
                                                     TRACE_DEFAULT_METHODWIDTH,
                                                     TRACE_DEFAULT_LINENOWIDTH,
                                                     TRACE_DEFAULT_MAXSIZE);
        } else {
          throw std::string("incompatible existing trace config");
        }
      }

      // Versions match, so just use this as our TraceOverall object, and increment the
      // use count
      spInstance->mUseCount++;
    }

    // Get access to the trace output stream.  This stream is held as a single per-process
    // global
    spTraceStream = new std::ofstream(spInstance->mTraceFile1, std::ios::binary|std::ios::app);
    if (spTraceStream->tellp() > spInstance->mMaxFileSize) {
      // First trace file is too big -- move to second file
      // If this is also too big, we will switch back after the first trace output
      delete spTraceStream;
      spTraceStream = new std::ofstream(spInstance->mTraceFile2, std::ios::binary|std::ios::app);
      spInstance->mWhich = 2;
    }
    sPid = GETPID();
  }
  return spInstance;
}

TraceOverall::TraceOverall(const char * filename1, const char * filename2,
                           int lvl, bool inc, int cw, int mw, int lw, long maxsize) :
  mMajorVersion(MAJORVERSION), mMinorVersion(MINORVERSION),
  mStructSize(sizeof(TraceOverall)), mUseCount(1),
  mTraceLevel(lvl), mIncludeTID(inc), mClassWidth(cw), mMethodWidth(mw), mLineWidth(lw),
  mMaxFileSize(maxsize), mWhich(1)
{
  strncpy(mTraceFile1, filename1, sizeof(mTraceFile1));
  strncpy(mTraceFile2, filename2, sizeof(mTraceFile2));
  mTraceFile1[sizeof(mTraceFile1)-1] = '\0';
  mTraceFile2[sizeof(mTraceFile2)-1] = '\0';
}

void TraceOverall::maybeSwitchFiles()
{
  if (spTraceStream->tellp() > mMaxFileSize) {
    delete spTraceStream;
    if (mWhich == 1) {
      spTraceStream = new std::ofstream(mTraceFile2, std::ios::binary|std::ios::trunc|std::ios::out);
      mWhich = 2;
    } else {
      spTraceStream = new std::ofstream(mTraceFile1, std::ios::binary|std::ios::trunc|std::ios::out);
      mWhich = 1;
    }
  }
}

// Destructor will never be called
TraceOverall::~TraceOverall()
{
  abort();
}

// Class with a single instance, in order to terminate trace at end-of-day
class TraceTerminator {
public:
  TraceTerminator() {}
  ~TraceTerminator();
};

static TraceTerminator sTerminator;

TraceTerminator::~TraceTerminator()
{
  // Close trace stream
  delete TraceOverall::spTraceStream;
  TraceOverall::spTraceStream = 0;

  // Close access to shared memory
  TraceLock scopeLock(TraceOverall::sLockFD);
  if (TraceOverall::spInstance == 0) return;
  TraceOverall::spInstance->mUseCount--;
  munmap(spSharedMem, sizeof(TraceOverall));
  spSharedMem = 0;
  close(sFileId);
  sFileId = -1;

  // If we were storing the trace config elsewhere, we could now do:
  // if (lastOut) unlink(SHMFILE);
  // However, we may as well just leave the config file in place
  TraceOverall::spInstance = 0;
}

#endif // if TRC_COMPILE_LEVEL <= MAX_TRACE_LEVEL
#endif // ifdef TRC_COMPILE_LEVEL
