// -*-c++-*-
#ifndef _CPPTRACE_H_
#define _CPPTRACE_H_
// C++ Tracing macros
//
// Copyright (c) David Drysdale 2000, 2006, 2010
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
// AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
// AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// Except as contained in this notice, the name of the author shall not be
// used in advertising or otherwise to promote the sale, use or other dealings
// in this Software without prior written authorization from the author.
// ----------------------------------------------------------------
//
// To use these, in each method you need to declare the method name.  There are
// two ways to do this.
//   1) #define TRCCLASSNAME at the top of each file to give the class name
//      that applies throughout, then use TRACE_FN("function_name") at the
//      top of each method. (If TRCCLASSNAME is undefined it defaults to the
//      name of the file).
//   2) Use TRACE_FQN("classname", "function_name") at the top of each method.
//
// If you are using these macros in multi-threaded code, you should
// #define MT_TRACE before including this file (and in the compilation of
// cpptrace.cpp).
//
// After this, you can use the following trace macros:
//  Level     Macro     Name       Purpose
//   -1   {}  N/A     Entry/Exit  Function entry & exit
//    0      TRC_DBG    Debug     Lowest level of trace, per-arm
//    1   .  TRC_NRM    Normal    Significant events
//    2   +  TRC_ALT    Alert     Unusual occurrence but should proceed OK
//    3   *  TRC_ERR    Error     Error occurrence, loss of function
//    4   !  TRC_ASSERT Assert    Internal error
//    5   @  TRC_LOG    Log       User-visible significant event
//
// Each tracing macro needs to look like a iostreams output statement, but
// can trace parameters etc.
//
// Putting this together, a sample file using this tracing might look like:
//
//  #define TRCCLASSNAME "Fred"
//  #include <cpptrace.h>
//  #include <Fred.h>
//
//  void Fred::doSomething(int a)
//  {
//    TRACE_FN("doSomething");
//    // or TRACE_FQN("Fred", "doSomething");
//
//    TRC_DBG("called with "<<a);
//    if (a==0) {
//      TRC_NRM("deal with zero case, mOtherField is "<<mOtherField);
//      doOtherThing(&a);
//    }
//    TRC_ASSERT((a!=0), "a still zero, mField is "<<mField);
//  }
//
// Finally, at link time you'll need to include
//  a) the cpptrace.o object, created from cpptrace.cpp (with TRC_COMPILE_LEVEL
//     set, and with MT_TRACE set appropriately)
//  b) the standard C++ library, -lstdc++

//! Highest trace level
#define MAX_TRACE_LEVEL 5
//! Lowest (most verbose) trace level
#define MIN_TRACE_LEVEL -1

#ifndef TRC_COMPILE_LEVEL
//! Level of tracing compiled in; include all tracing by default
#define TRC_COMPILE_LEVEL MIN_TRACE_LEVEL
#endif

// Expand the trace macros depending on the compiled-in trace level
#if (TRC_COMPILE_LEVEL <= 0)
#define TRC_DBG(A)            TRCX(0, ' ', A)
#endif
#if (TRC_COMPILE_LEVEL <= 1)
#define TRC_NRM(A)            TRCX(1, '.', A)
#endif
#if (TRC_COMPILE_LEVEL <= 2)
#define TRC_ALT(A)            TRCX(2, '+', A)
#endif
#if (TRC_COMPILE_LEVEL <= 3)
#define TRC_ERR(A)            TRCX(3, '*', A)
#endif
#if (TRC_COMPILE_LEVEL <= 4)
// We need C's stdlib.h for abort():
#include <cstdlib>
#define TRC_ABORT(A)          {TRCX(4, '!', A); std::cerr<<"Internal error: "<<A<<std::endl; abort();}
#define TRC_ASSERT(C, A)      if (!(C)) {TRC_ABORT(A);}
#endif
#if (TRC_COMPILE_LEVEL <= 5)
#define TRC_LOG(A)            TRCX(5, '@', A)
#endif

#if (TRC_COMPILE_LEVEL <= MAX_TRACE_LEVEL)

#ifndef TRCCLASSNAME
//! Class name for current file; use the filename if no class name is given
#define TRCCLASSNAME __FILE__
#endif

#include <unistd.h>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <string>
#include <ctime>

//! Overall trace control parameters are held in this class.  There is only one; however
//! we still want it to be a class rather than a namespace so that we can do initialization
//! and cleanup when the single global instance is created/destroyed.
class TraceOverall {
public:
  //! Retrieve one true instance (Singleton pattern)
  static TraceOverall* instance();
protected:
  //! The one true instance
  static TraceOverall* spInstance;
  friend class TraceTerminator;

public:
  void maybeSwitchFiles();
  int getMajorVersion();
  int getMinorVersion();
  int getUseCount();
  int getTraceLevel();
  bool getIncludeTID();
  int getClassWidth();
  int getMethodWidth();
  int getLineWidth();
  const char* getTraceFile1();
  const char* getTraceFile2();
  long getMaxFileSize();

  void setTraceLevel(int level);
  void setIncludeTID(bool inc);
  void setClassWidth(int width);
  void setMethodWidth(int width);
  void setLineWidth(int width);
  void setMaxFileSize(long size);

  TraceOverall(const char * filename1, const char * filename2,
               int lvl, bool inc, int cw, int mw, int lw, long maxsize);
  ~TraceOverall();
  TraceOverall(const TraceOverall&);            // make inaccessible
  TraceOverall& operator=(const TraceOverall&); // make inaccessible

private:
  // Shared data
  const int mMajorVersion; // Note: these version fields *must* be the first
  const int mMinorVersion; // two fields in the structure.
  const int mStructSize;   // Consistency check: same size?
  int      mUseCount;      // How many processes accessing this?
  int      mTraceLevel;    // Overall current runtime trace level
  bool     mIncludeTID;    // Include thread ID?
  int      mClassWidth;    // Field width for printing class name
  int      mMethodWidth;   // Field width for printing method name
  int      mLineWidth;     // Field width for printing line number
  char     mTraceFile1[256]; // Filenames of files to output trace to
  char     mTraceFile2[256];
  long     mMaxFileSize;   // Size at which to switch streams
  int      mWhich;         // Which are we currently outputting to?
public:
  // Per-process data which needs to be accessed everywhere
  static int sPid;         // Process ID
  static int sLockFD;      // FD for trace lock file
  static std::ofstream *spTraceStream; // Output stream
  static bool sInTrace;    // In mid-trace
};

// Accessor methods
inline int TraceOverall::getMajorVersion() {return mMajorVersion;}
inline int TraceOverall::getMinorVersion() {return mMinorVersion;}
inline int TraceOverall::getUseCount()     {return mUseCount;}
inline int TraceOverall::getTraceLevel()   {return mTraceLevel;}
inline bool TraceOverall::getIncludeTID()  {return mIncludeTID;}
inline int TraceOverall::getClassWidth()   {return mClassWidth;}
inline int TraceOverall::getMethodWidth()  {return mMethodWidth;}
inline int TraceOverall::getLineWidth()    {return mLineWidth;}
inline const char* TraceOverall::getTraceFile1() {return mTraceFile1;}
inline const char* TraceOverall::getTraceFile2() {return mTraceFile2;}
inline long TraceOverall::getMaxFileSize() {return mMaxFileSize;}

inline void TraceOverall::setTraceLevel(int level)  {mTraceLevel=level;}
inline void TraceOverall::setIncludeTID(bool inc)   {mIncludeTID=inc;}
inline void TraceOverall::setClassWidth(int width)  {mClassWidth=width;}
inline void TraceOverall::setMethodWidth(int width) {mMethodWidth=width;}
inline void TraceOverall::setLineWidth(int width)   {mLineWidth=width;}
inline void TraceOverall::setMaxFileSize(long size) {mMaxFileSize=size;}

// Convenience macros
#define SET_TRACE_LEVEL(n) TraceOverall::instance()->setTraceLevel(n)
#define GET_TRACE_LEVEL    TraceOverall::instance()->getTraceLevel()

#ifdef MT_TRACE
#include <lock.h>
extern CriticalSectionObject** sppTraceThreadLock;
#endif

// Class for acquiring a lock within a scope, exception safely
class TraceLock {
public:
  explicit TraceLock(int fd) : mLockFileDes(fd) {
    if (mLockFileDes == -1) return;
    if (lockf(mLockFileDes, F_LOCK, 0) != 0) throw std::string("failed to lock trace lock");
#ifdef MT_TRACE
    if (*sppTraceThreadLock) (*sppTraceThreadLock)->getLock();
#endif
  }
  ~TraceLock() {
    if (mLockFileDes == -1) return;
    if (lockf(mLockFileDes, F_ULOCK, 0) != 0) {
      // Failed to unlock; nothing we can do
    }
#ifdef MT_TRACE
    if (*sppTraceThreadLock) (*sppTraceThreadLock)->releaseLock();
#endif
  }
private:
  int mLockFileDes;
};

// Define a way of printing out date/time for tracing
inline std::ostream& operator<<(std::ostream &s, struct tm* t)
{
  return s<<std::setw(4)<<(t->tm_year+1900)<<"-"<<
            std::setw(2)<<std::setfill('0')<<(t->tm_mon+1)<<"-"<<
            std::setw(2)<<t->tm_mday<<" "<<
            std::setw(2)<<t->tm_hour<<":"<<
            std::setw(2)<<t->tm_min<<":"<<
            std::setw(2)<<t->tm_sec<<std::setfill(' ');
}

#ifdef MT_TRACE
#ifdef _WIN32
#include <windows.h>
#define THREAD_ID GetCurrentThreadId()
#else
#include <pthread.h>
#define THREAD_ID pthread_self()
#endif
#else
#define THREAD_ID ""
#endif

//! Master macro that emits a trace line
#define TRCXX(LVL, CLS, NM, A, LN, SEP)                                 \
  { TraceLock scopeLock(TraceOverall::sLockFD);                         \
    if ((!TraceOverall::instance()->sInTrace) &&                        \
        (TraceOverall::instance()->getTraceLevel() <= LVL)) {           \
      TraceOverall::instance()->sInTrace = true;                        \
      time_t now = time(0);                                             \
      TraceOverall::spTraceStream->seekp(0, std::ios::end);             \
      (*TraceOverall::spTraceStream) <<                                 \
        localtime(&now) << ' ' <<                                       \
        std::setw(5) << TraceOverall::sPid;                             \
      if (TraceOverall::instance()->getIncludeTID()) {                  \
        (*TraceOverall::spTraceStream) << ":" << THREAD_ID;             \
      }                                                                 \
      (*TraceOverall::spTraceStream) << ' ' <<                          \
        std::setw(TraceOverall::instance()->getClassWidth()) <<         \
        std::setiosflags(std::ios::right) <<                            \
        CLS <<                                                          \
        std::resetiosflags(std::ios::right) <<                          \
        "::" <<                                                         \
        std::setw(TraceOverall::instance()->getMethodWidth()) <<        \
        std::setiosflags(std::ios::left) <<                             \
        NM <<                                                           \
        std::resetiosflags(std::ios::left) <<                           \
        ' ' <<                                                          \
        std::setw(TraceOverall::instance()->getLineWidth()) <<          \
        LN <<                                                           \
        ' ' << SEP << ' ' << A <<                                       \
        std::endl << std::flush;                                        \
      TraceOverall::instance()->maybeSwitchFiles();                     \
      TraceOverall::instance()->sInTrace = false;                       \
    }                                                                   \
  }

//! Wrapper macro to emit a trace line; includes the current line number
#define TRCX(LVL, SEP, A)                                               \
  TRCXX(LVL, trcobj.mClassName, trcobj.mMethodName, A, __LINE__, SEP);

// Per-method information is kept in a TraceObject instance, so that its
// constructor and destructor can trace entry/exit.
class TraceObject {
public:
  TraceObject(const char* className, const char *name, int lineNo);
  ~TraceObject();
  const char *mMethodName;
  const char *mClassName;
private:
  TraceObject(const TraceObject&);            // make inaccessible
  TraceObject& operator=(const TraceObject&); // make inaccessible
};

inline TraceObject::TraceObject(const char* className, const char *name, int lineNo) :
  mMethodName(name),
  mClassName(className)
{
#if (TRC_COMPILE_LEVEL >= -1)
  TRCXX(-1, mClassName, mMethodName, "Entry", lineNo, '{');
#endif
}

inline TraceObject::~TraceObject()
{
#if (TRC_COMPILE_LEVEL >= -1)
  TRCXX(-1, mClassName, mMethodName, "Exit", 0, '}');
#endif
}

// The per-method TraceObject has a standard name in all methods
#define TRACE_FN(F)     TraceObject trcobj(TRCCLASSNAME, F, __LINE__)
#define TRACE_FQN(C,F)  TraceObject trcobj(C, F, __LINE__)

#else
#error "TRC_COMPILE_LEVEL defined but bigger than MAX_TRACE_LEVEL"
#endif // if (TRC_COMPILE_LEVEL <= MAX_TRACE_LEVEL)


// If trace macros are defined at this point, they should expand to nothing
#ifndef TRC_DBG
#define TRC_DBG(A)
#endif
#ifndef TRC_NRM
#define TRC_NRM(A)
#endif
#ifndef TRC_ALT
#define TRC_ALT(A)
#endif
#ifndef TRC_ERR
#define TRC_ERR(A)
#endif
#ifndef TRC_LOG
#define TRC_LOG(A)
#endif
#ifndef TRC_ASSERT
#define TRC_ASSERT(A, B)
#endif
#ifndef TRC_ABORT
#define TRC_ABORT(A)
#endif
#ifndef TRACE_FN
#define TRACE_FN(A)
#endif

#endif
