// Lock objects for UNIX and Win32 - implementation
//
// Copyright (c) David Drysdale 2000
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
// AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
// AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// Except as contained in this notice, the name of the author shall not be
// used in advertising or otherwise to promote the sale, use or other dealings
// in this Software without prior written authorization from the author.

#define TRCCLASSNAME "DebugLock"
#ifndef DEBUG
#define DEBUG
#endif
#include <cpptrace.h>

#include <lock.h>

#include <sstream>
#include <vector>
#include <queue>
#include <string>
#include <list>
#include <map>
#include <set>

// This class deals with tracking all of the current locks in the program.  It
// is accessed via a singleton getInstance() method.
class DebugLockTracker {
public:
  static DebugLockTracker *getInstance();
  void addCurrentLock(const std::string& lockname, const std::string& where);
  void removeCurrentLock(const std::string& lockname, const std::string& where);
  void dumpLockStatus();

protected:
  DebugLockTracker() : mDeadLockSeen(false) {}
  ~DebugLockTracker() {}
  bool checkLoops(); // look for loops in the lock hierarchy
  void visit(std::string k);

  // singleton instance pointer
  static DebugLockTracker* sInstance;
  // Protect access to the instance of this class against multi-threaded access
  // Note that we use the "raw" critical section object to prevent recursion
  static CriticalSectionObject sClassLock;
  // List of the locks currently held by the program. Each entry in this
  // list is a pair (lock name, location lock was acquired) -- for example
  // ("lockA", "fred.c:34").
  // TODO:@@@@ this list should be per-thread @@@@@
  std::list<std::pair<std::string, std::string>*> mCurrentLocks;
  // The adjacency list, which represents the locking sequences seen by
  // the program. For example, a thread that locks A, then B, then unlocks
  // B and then A will add entries (0, A) and (A, B) to this data structure
  std::map<std::string, std::set<std::string> > mAdjacencyList;
  // We only want to scream about the first deadlock that gets
  // detected, since it will keep on being re-detected
  bool mDeadLockSeen;
};

// Single global needed for the lock tracker.  Note that this pointer will be
// initialized before any constructors are called (since it a pointer is a
// built-in type)
DebugLockTracker* DebugLockTracker::sInstance(0);
// Function to return the singleton instance
DebugLockTracker* DebugLockTracker::getInstance()
{
  if (!sInstance) sInstance = new DebugLockTracker();
  return sInstance;
}
// Define the global critical section object which protects multi-threaded
// access to the DebugLockTracker instance.  This object will have its
// constructor called at start-of-day -- i.e. while the program is still
// single threaded.
CriticalSectionObject DebugLockTracker::sClassLock;

// Method to dump out the adjacency list.
void DebugLockTracker::dumpLockStatus() {
  TRACE_FN("dumpLockStatus");
  CSOAcquisition acquired(sClassLock);
  TRC_NRM("Locking sequences observed:");
  for (std::map<std::string, std::set<std::string> >::const_iterator ii = mAdjacencyList.begin();
       ii != mAdjacencyList.end(); ii++) {
    const std::set<std::string>& connects((*ii).second);
    for (std::set<std::string>::const_iterator jj = connects.begin();
         jj != connects.end(); jj++) {
      TRC_NRM((*ii).first<<"->"<< (*jj));
    }
  }
  checkLoops();
}

bool DebugLockTracker::checkLoops()
{
  TRACE_FN("checkLoops");

  // build an empty map <string, int> for each entry in the adjacency list.  The
  // aim is to generate an ordering which respects the partial ordering of the
  // locks.
  std::map<std::string, int> val;
  int id(0);
  std::queue<std::string> q; // queue of nodes seen at previous level

  // pre-initialize the numbering to zeroes
  for (std::map<std::string, std::set<std::string> >::const_iterator kk = mAdjacencyList.begin();
       kk != mAdjacencyList.end(); kk++) {
    val[(*kk).first] = 0;
  }

  // catch the case where we haven't got any locks
  if (mAdjacencyList.size() == 0) return(false);

  // can reach everything from "0".  Breadth first search
  std::string k = (*mAdjacencyList.begin()).first;

  bool foundLoop = false;
  q.push(k);
  do {
    k = q.front(); q.pop();
    id++;
    val[k] = id;
    TRC_DBG("node "<<k<<" gets id "<<id);
    for (std::set<std::string>::const_iterator ii=mAdjacencyList[k].begin();
         ii != mAdjacencyList[k].end(); ii++) {
      if (val[*ii]==0) {
        q.push(*ii);
        val[*ii]=-1;
        TRC_DBG("node "<<*ii<<" gets id -1 for now");
      } else if (val[*ii] != -1) {
        TRC_ALT("seen node "<<*ii<<" more than once");
        foundLoop = true;
      }
    }
  } while (q.size() != 0);
  return(foundLoop);
}

void DebugLockTracker::addCurrentLock(const std::string& lockname, const std::string& where)
{
  TRACE_FN("addCurrentLock");
  CSOAcquisition acquired(sClassLock);

  // Add this pair <prevlock, thislock> to the adjacency list
  std::string left((mCurrentLocks.size()==0) ? "0" : (*mCurrentLocks.rbegin())->first);
  std::set<std::string> &connects = mAdjacencyList[left];
  connects.insert(lockname);

  // Look for possible loops in the adjacency list on the fly
  if (checkLoops()) {
    if (!mDeadLockSeen) {
      mDeadLockSeen = true; // only report the first deadlock as an error at runtime
      TRC_ERR("Potential deadlock seen on call to lock "<<lockname<<" from "<<where);
    } else {
      // we will keep detecting the same deadlock, so lower the trace level
      TRC_NRM("Potential deadlock seen again on call to lock "<<lockname<<" from "<<where);
    }
  }

  // Now update the list of currently held locks
  mCurrentLocks.push_back(new std::pair<std::string,std::string>(lockname, where));
  TRC_DBG("Locks held:");
  for (std::list<std::pair<std::string, std::string>*>::const_iterator ii = mCurrentLocks.begin();
       ii != mCurrentLocks.end(); ii++) {
    TRC_DBG("  "<<(*ii)->first<<", locked at "<<(*ii)->second);
  }
}

void DebugLockTracker::removeCurrentLock(const std::string& lockname, const std::string& where)
{
  TRACE_FN("removeCurrentLock");
  CSOAcquisition acquired(sClassLock);

  // search through the list of current locks to find this one. We expect to find it
  // as the last entry in the list -- if it is not, then this indicates a problem
  bool found=false;
  for (std::list<std::pair<std::string, std::string>*>::const_iterator ii = mCurrentLocks.begin();
       ii != mCurrentLocks.end(); ii++) {
    if ((*ii)->first == lockname) { // found it
      if (*ii != *mCurrentLocks.rbegin()) { // but not at the end of the list
        TRC_ERR("Releasing lock "<<lockname<<" at "<<where);
        TRC_ERR("Last lock was of "<<(*mCurrentLocks.rbegin())->first<<
                " from "<<(*mCurrentLocks.rbegin())->second);
      }
      // remove the lock from the list
      mCurrentLocks.remove(*ii);
      delete(*ii);
      found=true;
      break;
    }
  }
  if (!found) {
    TRC_ERR("lock "<<lockname<<" being released from "<<where<<" not found");
  }
  TRC_DBG("Locks held:");
  for (std::list<std::pair<std::string, std::string>*>::const_iterator ii = mCurrentLocks.begin();
       ii != mCurrentLocks.end(); ii++) {
    TRC_DBG("  "<<(*ii)->first<<", locked at "<<(*ii)->second);
  }
}

// Constructor for a debug CriticalSectionObject records an identifier for the lock, which
// contains the name of the variable plus where it was declared.
DebugCriticalSectionObject::DebugCriticalSectionObject(const char* name, const char* filename, int line) :
  mLockCount(0)
{
  TRACE_FN("DebugCriticalSectionObject");

  std::stringstream ss;
  ss << name << "(" << filename << ":" << line << ")";
  mLabel = ss.str();
  TRC_NRM("New critical section "<<mLabel);
}
DebugCriticalSectionObject::~DebugCriticalSectionObject()
{
  TRACE_FN("~DebugCriticalSectionObject");
  TRC_NRM("Remove critical section "<<mLabel);
}
void DebugCriticalSectionObject::getLock(const char* filename, int line)
{
  TRACE_FN("getLock");
  char tempArray[1024];
  sprintf(tempArray, "%s:%d", filename, line);
  std::string whereLabel(tempArray);
  DebugLockTracker::getInstance()->addCurrentLock(mLabel, whereLabel);
  CriticalSectionObject::getLock();
  mLockCount++; // Note: do this inside the lock
}
void DebugCriticalSectionObject::releaseLock(const char* filename, int line)
{
  TRACE_FN("releaseLock");
  char tempArray[1024];
  sprintf(tempArray, "%s:%d", filename, line);
  std::string whereLabel(tempArray);
  DebugLockTracker::getInstance()->removeCurrentLock(mLabel, whereLabel);
  mLockCount--; // Note: do this inside the lock
  if (mLockCount < 0) {
    TRC_ERR("lock count for "<<mLabel<<" gone negative ("<<mLockCount<<") at "<<whereLabel);
  }
  CriticalSectionObject::releaseLock();
}

// Create a class in order to perform dumping of the locking sequences at end-of-day
class DebugLockEdgeDumper {
public:
  DebugLockEdgeDumper() {}
  ~DebugLockEdgeDumper() {DebugLockTracker::getInstance()->dumpLockStatus();}
};
// Single global instance, whose destructor prints out the locking
// hierarchy seen in a run
static DebugLockEdgeDumper sOne;
