// -*-c++-*-
#ifndef _VERSION_H
#define _VERSION_H
// Functionality for recording the version information for a source code file
//
// Copyright (c) 2005-2006,2010 David Drysdale
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
// AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
// AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// Except as contained in this notice, the name of the author shall not be
// used in advertising or otherwise to promote the sale, use or other dealings
// in this Software without prior written authorization from the author.
// ----------------------------------------------------------------
//
// To use this, just include a line like:
//    RCSID(version_h, "$Id: version.h,v 1.6 2010-05-24 16:46:45 dmd Exp $");
// at the top of a source file.  The second argument will be expanded by the
// version control system to include the file's current revision information.
//
// This macro will include an version stamp in the compiled version of
// the file and this versioning information will be automatically registered
// at start of day.
//
// This code is not thread-safe, but for normal use this should make no
// difference -- all of the functionality occurs at start-of-day, during
// global object construction, which happens before any additional threads
// have been spawned.

#include <map>
#include <string>
#include <ostream>
#include <sstream>
#include <stdexcept>

//! Macro to include a static local variable with version information
#if defined(__GNUC__)
#include <sys/cdefs.h>
#define	IDSTRING(name,string) static const char* name __attribute__((__unused__)) = string
#else
#define IDSTRING(name,string) static const char* name = string
#endif

//! Collection of source code file version information
class FileRegister {
public:
  //! Add a file to the register.
  //! @param[in] name Name of file to add
  //! @param[in] version Version information about file
  //! @param[in] info Verbose information about the file to add
  //! @param[in] date Build date for file
  //! @param[in] time Build time for file
  void addFile(const std::string& name,
               const std::string& version,
               const std::string& info,
               const char* date,
               const char* time) {
    // Check for existing version information
    std::map<std::string, std::string>::const_iterator existing_version = version_.find(name);
    if ((existing_version != version_.end()) &&
        (existing_version->second != version)) {
      std::stringstream ss;
      ss << "Two different versions of " << name << ":" << std::endl;
      ss << "  " << version_[name] << " info " << info_[name] << " built " << date_[name] << " " << time_[name] << std::endl;
      ss << "  " << version << " info " << info << " built " << date << " " << time << std::endl;
      throw std::range_error(ss.str());
    }
    version_[name] = version;
    info_[name] = info;
    date_[name] = date;
    time_[name] = time;
  }
  //! Return (multi-line) string with all version information in it
  std::string to_str();
  //! Return string containing largest version information seen
  //! This assumes version information of the form "$Revision: 1.6 $"
  //! (this is only really useful for Subversion-controlled files)
  int max_version();
  //! Find single instance of class (Singleton pattern).
  //! @return The one true instance
  static FileRegister* instance() {
    if (spInstance == 0) {
      spInstance = new FileRegister;
    }
    return spInstance;
  }
protected:
  //! Static instance pointer for the Singleton pattern
  static FileRegister* spInstance;
public:
  //! Version information for the files
  std::map<std::string, std::string> version_;
  //! Verbose information for the files
  std::map<std::string, std::string> info_;
  //! Build date information for the files
  std::map<std::string, const char*> date_;
  //! Build time information for the files
  std::map<std::string, const char*> time_;
};

inline std::ostream& operator<<(std::ostream& s, FileRegister& r)
{
  return s << r.to_str();
}

//! Class to trigger registration on construction
struct FileRegistrar {
  //! Constructor
  //! @param[in] name Identifier of file to add
  //! @param[in] version Version information about file
  //! @param[in] info Verbose information about the file to add
  //! @param[in] date Build date for file
  //! @param[in] time Build time for file
  FileRegistrar(const char* name,
                const char* version,
                const char* info,
                const char* date,
                const char* time) {
    FileRegister::instance()->addFile(name, version, info, date, time);
  }
};

//! Main version stamp macro.  This does two things: firstly it embeds
//! a string in the object file with version information.  Secondly
//! it includes a file-scope FileRegistrar global object.  When this
//! global object is constructed at start of day, it registers the
//! file's version information with the one true FileRegister.
#ifndef RCSID
//! Version information generation
#define RCSID(prefix, v, i)           \
  IDSTRING(version_ ## prefix, v i);  \
  static FileRegistrar add_ ## prefix(#prefix, v, i, __DATE__, __TIME__)
#endif
#ifndef SVNID
#define SVNID(prefix, v) \
  IDSTRING(version_ ## prefix, v " $Filename: " __FILE__ " $");  \
  static FileRegistrar add_ ## prefix(#prefix, v, __FILE__, __DATE__, __TIME__)
#endif

//! Version information for this file itself
//RCSID(version_h, "$Revision: 1.6 $", "$RCSfile: version.h,v $");
SVNID(version_h, "$Revision: 1.6 $");

#endif
